// RC522 UART interface
// Written by Boomer48: June 2019


#include "LCD_4Bit.h"

// Input switches
#define Init_Sw 11 // manual re-init of tag
#define Inc_Sw 12 // Increment tag credit amount

#define Trail_Blk 7 // ID of Trailer Block (holds keys)
#define Data_Blk 6 // ID of Data Block (holds credits)
#define Bkup_Blk 5 // ID of Backup Data Block (holds credits)
#define Firm_Vers 0x92 // Module firware version (0x91 or 0x92)

#define Def_Key1 0xFF // first byte of default key
#define Def_Key2 0xFF
#define Def_Key3 0xFF
#define Def_Key4 0xFF
#define Def_Key5 0xFF
#define Def_Key6 0xFF // last byte of default key

#define My_Key1 0x01 // first byte of new key
#define My_Key2 0x02
#define My_Key3 0x00
#define My_Key4 0x01
#define My_Key5 0x04
#define My_Key6 0x08 // last byte of new key

// Set for increment amount (packed BCD)
#define Inc1        0x00 // 10/1
#define Inc100      0x00 // 1000/100
#define Inc10k      0x00 // 100/10 thousand
#define Inc1M       0x02 // 10/1 million

// Set for decrement amount (packed BCD)
#define Dec1        0x00 // 10/1
#define Dec100      0x00 // 1000/100
#define Dec10k      0x50 // 100/10 thousand
#define Dec1M       0x00 // 10/1 million

byte temp;
byte Tag_Blk_Num; // Tag Block Number
byte UID1;
byte UID2;
byte UID3;
byte UID4;
byte UID_BCC; // Block Check Character (XOR of UID)
byte FIFO_Data; // temporary holder for FIFO byte
//byte Line_Brk; // LCD line break after this value
byte CRC_L; // buffer for CRC low value
byte CRC_H; // buffer for CRC high value
byte Exp_Resp; // expected response to module commands
byte Cmd_Resp; // response received
byte ATQA_L; // Tag Answer-To-Request-A (low)
byte ATQA_H; // Tag Answer-To-Request-A (high)
byte Junk_Num = 0; // junk number for write to tag
byte Error_Flg;
byte Millions; // Packed BCD (millions and 10 millions)
byte Ten_Thou; // Packed BCD (10 thousands and 100 thousands)
byte Hundreds; // Packed BCD (hundreds and thousands)
byte Ones; // Packed BCD (ones and tens)
byte Carry; // carry flag from DAA routine
byte Borrow; // borrow flag from DAS routine
byte Comp_1; // 1's complement credit amount LS byte
byte Comp_2;
byte Comp_3;
byte Comp_4; // 1's complement credit amount MS byte
byte Try_Def; // if scan fails then try default key
byte Reinit_Tag; // tag data corrupt so try reinit
byte Bad_Credit; // flag set if credit data mismatch

char ASCII[] = "0123456789ABCDEF";

void setup()
{
  pinMode(Init_Sw, INPUT_PULLUP);
  pinMode(Inc_Sw, INPUT_PULLUP);
  LCD_setup();
  Serial.begin(9600);
Init_522:
  Error_Flg = false;
  Flush_RX();
  Startup();

  if (Error_Flg == true)
    goto Init_522;
}

void loop()
{
Scan_Loop:
  Error_Flg = false;
  Tag_Blk_Num = Data_Blk;

  // Need to clear the crypto flag before each scan
  Clr_Crypto();

  // Find a tag
  Scan_Tag();

  if (Error_Flg == true)
    goto Scan_Loop;

  Disp_UID(); // comment out if not needed

  // Found a tag, now select it
  Select_Tag();

  if (Error_Flg == true)
    goto Scan_Loop;

  Select_Msg(); // comment out if not needed

  // Authenticate for reads/writes
  Auth_Tag();

  if (Error_Flg == true)
    goto Scan_Loop;

  Auth_Msg(); // comment out if not needed

  if (Error_Flg == true)
    goto Scan_Loop;

  // Reinit tag based on manual command, failed key access, or corrupt data
  if ((digitalRead(Init_Sw) == LOW) || (Try_Def == true) || (Reinit_Tag == true)) {
    Init_Tag();
    goto Scan_Loop;
  }

  // Read from tag
  Read_Tag();

  if (Error_Flg == true)
    goto Scan_Loop;

  Get_Credit(); // get stored credit amount

  if (Bad_Credit == true) {
    Match_Fail();
    goto Scan_Loop;
  }

  Disp_Credit(); // display credit amount

  if (digitalRead(Inc_Sw) == LOW)
    Inc_Credit(); // increment credits
  else
    Dec_Credit(); // decrement credits

  if (Error_Flg == true)
    goto Scan_Loop;

  Done_Msg();
}

// -----------------------------------
// Receive Buffer Utilities (debug)
// -----------------------------------

void Flush_RX() {
  while (Serial.available() > 0)
    Serial.read(); // clear receive buffer
}

void Get_FIFO()
{
  Read_FIFO();
  Read_FIFO();
  Read_FIFO();
  Read_FIFO();
  Read_FIFO();
  Read_FIFO();
  Read_FIFO();
  Read_FIFO();
  Read_FIFO();
  Read_FIFO();
  Read_FIFO();
  Read_FIFO();
  Read_FIFO();
  Read_FIFO();
  Read_FIFO();
  Read_FIFO();
}

void Make_ASCII(byte BCD_Byte) {
  Send_Byte(ASCII[BCD_Byte >> 4]); // display upper nibble first
  Send_Byte(ASCII[BCD_Byte & 0x0F]);
}

void Disp_Data()
{
  byte i;

  LCD_Line1();

  for (i = 0; i < 8; i++)
    Make_ASCII(Serial.read());

  LCD_Line2();

  for (i = 0; i < 8; i++)
    Make_ASCII(Serial.read());
}

// -----------------------------------
// Data conversion routines
// -----------------------------------

// Supress leading zeros in credits display
void Find_First() {
  if ((Millions & 0xF0) != 0)
    goto Disp_10M;
  else if ((Millions & 0x0F) != 0)
    goto Disp_1M;
  else if ((Ten_Thou & 0xF0) != 0)
    goto Disp_100k;
  else if ((Ten_Thou & 0x0F) != 0)
    goto Disp_10k;
  else if ((Hundreds & 0xF0) != 0)
    goto Disp_1k;
  else if ((Hundreds & 0x0F) != 0)
    goto Disp_100;
  else if ((Ones & 0xF0) != 0)
    goto Disp_10;
  else
    goto Disp_1;

Disp_10M:
  Send_Byte(ASCII[Millions >> 4]);
Disp_1M:
  Send_Byte(ASCII[Millions & 0x0F]);
  Send_Byte(',');
Disp_100k:
  Send_Byte(ASCII[Ten_Thou >> 4]);
Disp_10k:
  Send_Byte(ASCII[Ten_Thou & 0x0F]);
Disp_1k:
  Send_Byte(ASCII[Hundreds >> 4]);
  Send_Byte(',');
Disp_100:
  Send_Byte(ASCII[Hundreds & 0x0F]);
Disp_10:
  Send_Byte(ASCII[Ones >> 4]);
Disp_1:
  Send_Byte(ASCII[Ones & 0x0F]);
}

// ---------------------------------------------------
// 522 Module Messages
// ---------------------------------------------------

void Startup()
{
  Dummy_Data(); // returns either 0x00 or nothing
  Soft_Reset();
  ASK100_Cmd();
  CRC_6363();
  Antenna_On();
  FW_Version();
}

// Board may power up thinking a command without data has been received.
// Neither a SW nor a HW reset corrects this so send a dummy byte.  If
// the byte is interpreted as a command it will be harmless.
void Dummy_Data()
{
  Serial.write(0x80); // read reserved register 0
  delay(2);
}

// Reset the module
void Soft_Reset()
{
  Serial.write(0x01); // write CommandReg
  Serial.write(0x0F); // reset 522
  delay(10);
}

// Set Amplitude Shift Key modulation to 100%
void ASK100_Cmd()
{
  Flush_RX();
  Exp_Resp = 0x15; // TxASKReg
  Serial.write(Exp_Resp);
  Serial.write(0x40); // 100% ASK
  delay(2);
  Chk_Resp();
}

// Set CRC calculation preset value to 6363h
void CRC_6363()
{
  Flush_RX();
  Exp_Resp = 0x11; // ModeReg
  Serial.write(Exp_Resp);
  Serial.write(0x3D); // CRC preset = 6363h
  delay(2);
  Chk_Resp();
}

// Turn antenna on (0x03).  Hardcoded to included the default
// value of 0x80 (TX2 signal inverted when TX2 enabled).
void Antenna_On()
{
  Flush_RX();
  Exp_Resp = 0x14; // write TxControlReg
  Serial.write(Exp_Resp);
  Serial.write(0x83); // Antenna On, RX2 inverted
  delay(2);
  Chk_Resp();
}

// Read module firmware version (0x91 or 0x92)
void FW_Version()
{
  Flush_RX();
  Serial.write(0xB7); // read VersionReg
  delay(2);
  Exp_Resp = Firm_Vers;
  Chk_Resp();
}

// Read the command register (results vary)
void Read_Cmd()
{
  Serial.write(0x81); // read CommandReg
  delay(2);
}

// Clear current command
void Idle_Cmd()
{
  Flush_RX();
  Exp_Resp = 0x01; // write CommandReg
  Serial.write(Exp_Resp);
  Serial.write(0x00); // set to idle
  delay(2);
  Chk_Resp();
}

// Mifare defines REQA and WUPA as 7-bit commands
void Bits7_Cmd()
{
  Flush_RX();
  Exp_Resp = 0x0D; // write BitFramingReg
  Serial.write(Exp_Resp);
  Serial.write(0x07); // use 7 bits of last byte in FIFO
  delay(2);
  Chk_Resp();
}

// All Mifare commands except REQA and WUPA are 8 - bits
void Bits8_Cmd()
{
  Flush_RX();
  Exp_Resp = 0x0D; // write BitFramingReg
  Serial.write(Exp_Resp);
  Serial.write(0x00); // use all 8 bits of FIFO bytes
  delay(2);
  Chk_Resp();
}

// Number of bytes in the response from the tag
void FIFO_Level()
{
  Serial.write(0x8A); // read FIFOLevelReg
  delay(2);
}

// Zero out FIFO byte count
void Flush_FIFO()
{
  Flush_RX();
  Exp_Resp = 0x0A; // write FIFOLevelReg
  Serial.write(Exp_Resp);
  Serial.write(0x80); // resets FIFO buffer pointer
  delay(2);
  Chk_Resp();
}

// Commands and data sent to the tag must be written to the FIFO.
// Byte to be written is assumed to be in the W register.
void Write_FIFO(byte FIFO_Byte)
{
  Flush_RX();
  Exp_Resp = 0x09; // write FIFODataReg
  Serial.write(Exp_Resp);
  Serial.write(FIFO_Byte);
  delay(2);
  Chk_Resp();
}

// Transfer a byte from the FIFO to the serial port
void Read_FIFO()
{
  Serial.write(0x89); // read FIFODataReg
  delay(2);
}

// Transmit bytes in the FIFO to the tag then turn on the receiver
void Transceive()
{
  Serial.write(0x01); // write CommandReg
  Serial.write(0x0C); // transmit FIFO data then activate receiver
  delay(2);
  Flush_RX();
  Serial.write(0x8D); // read BitFramingReg

  while (Serial.available() == 0)
    delay(1);

  Serial.write(0x0D); // write BitFramingReg
  temp = (Serial.read() | 0x80); // start transmission
  Serial.write(temp);
  delay(10);
}

// Authenticate the tag
void Auth_Cmd()
{
  Flush_RX();
  Exp_Resp = 0x01; // write CommandReg
  Serial.write(Exp_Resp);
  Serial.write(0x0E); // authenticate tag
  delay(10);
  Chk_Resp();
}

// Calculate the CRC for the bytes in the FIFO
// NOTE: This resets the FIFO_Level to zero
void Calc_CRC()
{
  Flush_RX();
  Exp_Resp = 0x01; // write CommandReg
  Serial.write(Exp_Resp);
  Serial.write(0x03); // calculate CRC for FIFO data
  delay(2);
  Chk_Resp();
}

// Read the high byte of the CRC
void Read_CRC_H()
{
  Serial.write(0xA1); // read CRCResultReg (high)
  delay(2);
}

// Read the low byte of the CRC
void Read_CRC_L()
{
  Serial.write(0xA2); // read CRCResultReg (low)
  delay(2);
}

// Read Staus Register 1
void Read_Stat1()
{
  Serial.write(0x87); // read Status1Reg
  delay(2);
}

// Read Status Register 2
void Read_Stat2()
{
  Serial.write(0x88); // read Status2Reg
  delay(2);
}

// Read the Error Flag register
void Read_Error()
{
  Serial.write(0x86); // read ErrorReg
  delay(2);
}

// Clears Status Register 2 which includes the Crypto_On flag
// Crypto_On gets set when a tag is authenticated and must be
// cleared before another scan is attempted.
void Clr_Crypto()
{
  Flush_RX();
  Exp_Resp = 0x08; // write Status2Reg
  Serial.write(Exp_Resp);
  Serial.write(0x00); // clear crypto on flag
  delay(2);
  Chk_Resp();
}

// Utility to check for a valid command response.
// Writes to a register will return an echo of the register number.
// Reads of a register may return multiple bytes.
void Chk_Resp()
{
  delay(10); // wait for serial port response

  if (Serial.available() == 0)
    No_Resp();
  else
  {
    Cmd_Resp = Serial.read();
    if (Cmd_Resp != Exp_Resp)
      Bad_Resp();
  }
}

// ---------------------------------------------------
// RFID Messages
// ---------------------------------------------------
// Sending commands to the card/tag requires putting the information into the
// MFRC522 module FIFO.  First the command, second the data (if any), last the
// checksum.  The MFRC522 module can be commanded to calculate the checksums
// which is then read out so it can be written to the FIFO.  Lastly, a Transceive
// or Authenticate command is sent.

// Look for a tag
void Scan_Tag()
{
Start_Scan:
  Scan_Msg();
  Idle_Cmd();
  Flush_FIFO();
  Write_FIFO(0x26); // REQA
  Bits7_Cmd();
  Transceive();
  Flush_RX();
  FIFO_Level();
  temp = Serial.read();

  if (temp == 0)
    goto Start_Scan; // did not receive tag type

  if (temp != 0x02)
    Bad_Resp();

  Read_FIFO();
  Read_FIFO();
  ATQA_L = Serial.read();
  ATQA_H = Serial.read();
  // Add check for tag type here if desired.
  // Mifare Classic (1k) = 0004h
  Get_UID();
}

void Get_UID()
{
  Idle_Cmd();
  Flush_FIFO();
  Write_FIFO(0x93); // CLI
  Write_FIFO(0x20); // Anticollision
  Bits8_Cmd();
  Transceive();
  Flush_RX();
  Read_FIFO();
  Read_FIFO();
  Read_FIFO();
  Read_FIFO();
  UID1 = Serial.read();
  UID2 = Serial.read();
  UID3 = Serial.read();
  UID4 = Serial.read();
}

void Select_Tag()
{
  Idle_Cmd();
  Flush_FIFO();
  Write_FIFO(0x93); // CL1
  Write_FIFO(0x70); // Select
  UID_BCC = UID1;
  Write_FIFO(UID1);
  UID_BCC = UID_BCC ^ UID2;
  Write_FIFO(UID2);
  UID_BCC = UID_BCC ^ UID3;
  Write_FIFO(UID3);
  UID_BCC = UID_BCC ^ UID4;
  Write_FIFO(UID4);
  Write_FIFO(UID_BCC);
  Calc_CRC();
  Flush_RX();
  Read_CRC_H();
  Read_CRC_L();
  delay(10);
  CRC_H = Serial.read();
  CRC_L = Serial.read();
  // Calc_CRC flushes the FIFO buffer so need to reload
  Idle_Cmd();
  Write_FIFO(0x93); // CL1
  Write_FIFO(0x70); // Select
  UID_BCC = UID1;
  Write_FIFO(UID1);
  UID_BCC = UID_BCC ^ UID2;
  Write_FIFO(UID2);
  UID_BCC = UID_BCC ^ UID3;
  Write_FIFO(UID3);
  UID_BCC = UID_BCC ^ UID4;
  Write_FIFO(UID4);
  Write_FIFO(UID_BCC);
  Write_FIFO(CRC_L);
  Write_FIFO(CRC_H);
  Transceive();
  Flush_RX();
  FIFO_Level();
  Exp_Resp = 0x03; // SAK + two byte checksum
  Chk_Resp();
  Flush_RX();
  Read_FIFO();
  Exp_Resp = 0x08; // SAK (Select Acknowledge)
  Chk_Resp();
}

// Authenticate tag data sector using the defined Trailer block ID
void Auth_Tag()
{
  Idle_Cmd();
  Flush_FIFO();
  Write_FIFO(0x60); // authenticate with Key A
  Write_FIFO(Trail_Blk); // Trailer block for data sector

  if (Try_Def == true) {
    Write_FIFO(Def_Key1);
    Write_FIFO(Def_Key2);
    Write_FIFO(Def_Key3);
    Write_FIFO(Def_Key4);
    Write_FIFO(Def_Key5);
    Write_FIFO(Def_Key6); // end Key A
  }
  else {
    Write_FIFO(My_Key1);
    Write_FIFO(My_Key2);
    Write_FIFO(My_Key3);
    Write_FIFO(My_Key4);
    Write_FIFO(My_Key5);
    Write_FIFO(My_Key6); // end Key A
  }

  Write_FIFO(UID1); // tag UID (4 bytes)
  Write_FIFO(UID2);
  Write_FIFO(UID3);
  Write_FIFO(UID4); // end tag UID
  Auth_Cmd();
  Flush_RX();
  Read_Stat2();
  delay(10);
  temp = Serial.read();
  if (bitRead(temp, 3) == 0) // Crypto On flag
    Bad_Scan(); // tag failed authentication
}

// Read tag data from the defined data block of the authorized sector.
// Data bytes (16) + two byte checksum will be returned in the FIFO.
void Read_Tag()
{
  Idle_Cmd();
  Flush_FIFO();
  Write_FIFO(0x30); // Mifare read command
  Write_FIFO(Tag_Blk_Num);
  Calc_CRC();
  Flush_RX();
  Read_CRC_H();
  Read_CRC_L();
  delay(10);
  CRC_H = Serial.read();
  CRC_L = Serial.read();
  // Calc_CRC flushes the FIFO buffer so need to reload
  Idle_Cmd();
  Write_FIFO(0x30); // Mifare read command
  Write_FIFO(Tag_Blk_Num);
  Write_FIFO(CRC_L);
  Write_FIFO(CRC_H);
  Transceive();
  Flush_RX();
  FIFO_Level();
  delay(10);
  temp = Serial.read();
  if (temp != 18) // 16 data bytes + 2 byte checksum
    Bad_Read();
}

// Initiates a write to the tag
void Write_Tag()
{
  Idle_Cmd();
  Flush_FIFO();
  Write_FIFO(0xA0); // write command
  Write_FIFO(Tag_Blk_Num);
  Calc_CRC();
  Flush_RX();
  Read_CRC_H();
  Read_CRC_L();
  delay(10);
  CRC_H = Serial.read();
  CRC_L = Serial.read();
  // Calc_CRC flushes the FIFO buffer so need to reload
  Idle_Cmd();
  Flush_FIFO();
  Write_FIFO(0xA0); // write command
  Write_FIFO(Tag_Blk_Num);
  Write_FIFO(CRC_L);
  Write_FIFO(CRC_H);
  Transceive();
  Flush_RX();
  Read_FIFO();
  delay(10);
  temp = Serial.read();
  if (temp != 10) // acknowledge
    Bad_Write();
}

// Sends credit data as part of the Write_Tag command.
// Formatted as 4 bytes credit, 4 bytes 1's complement of credit,
// 8 bytes of zeros.  Called once for the data block and once for
// the backup block.
void Send_Data()
{
  Idle_Cmd();
  Flush_FIFO();
  Write_FIFO(Millions);
  Write_FIFO(Ten_Thou);
  Write_FIFO(Hundreds);
  Write_FIFO(Ones);
  Write_FIFO(~Millions);
  Write_FIFO(~Ten_Thou);
  Write_FIFO(~Hundreds);
  Write_FIFO(~Ones);
  Write_FIFO(0x00);
  Write_FIFO(0x00);
  Write_FIFO(0x00);
  Write_FIFO(0x00);
  Write_FIFO(0x00);
  Write_FIFO(0x00);
  Write_FIFO(0x00);
  Write_FIFO(0x00);
  Calc_CRC();
  Flush_RX();
  Read_CRC_H();
  Read_CRC_L();
  delay(10);
  CRC_H = Serial.read();
  CRC_L = Serial.read();
  // Calc_CRC flushes the FIFO buffer so need to reload
  Idle_Cmd();
  Write_FIFO(Millions);
  Write_FIFO(Ten_Thou);
  Write_FIFO(Hundreds);
  Write_FIFO(Ones);
  Write_FIFO(~Millions);
  Write_FIFO(~Ten_Thou);
  Write_FIFO(~Hundreds);
  Write_FIFO(~Ones);
  Write_FIFO(0x00);
  Write_FIFO(0x00);
  Write_FIFO(0x00);
  Write_FIFO(0x00);
  Write_FIFO(0x00);
  Write_FIFO(0x00);
  Write_FIFO(0x00);
  Write_FIFO(0x00);
  Write_FIFO(CRC_L);
  Write_FIFO(CRC_H);
  Transceive();
  Flush_RX();
  delay(10);
  Read_FIFO();
  temp = Serial.read();
  if (temp != 10) // acknowledge
    Bad_Write();
}

// Sends key data as part of the Write_Tag command.
// Formated as 6 bytes Key A, 3 bytes write access flags,
// 1 byte Don't Care, 6 bytes Key B.
void Send_Keys()
{
  Idle_Cmd();
  Flush_FIFO();
  Write_FIFO(My_Key1);
  Write_FIFO(My_Key2);
  Write_FIFO(My_Key3);
  Write_FIFO(My_Key4);
  Write_FIFO(My_Key5);
  Write_FIFO(My_Key6); // end key A
  Write_FIFO(0xFF); // write access flags
  Write_FIFO(0x07);
  Write_FIFO(0x80);
  Write_FIFO(0x69); // don't care byte
  Write_FIFO(My_Key1);
  Write_FIFO(My_Key2);
  Write_FIFO(My_Key3);
  Write_FIFO(My_Key4);
  Write_FIFO(My_Key5);
  Write_FIFO(My_Key6); // end key B
  Calc_CRC();
  Flush_RX();
  Read_CRC_H();
  Read_CRC_L();
  delay(10);
  CRC_H = Serial.read();
  CRC_L = Serial.read();
  // Calc_CRC flushes the FIFO buffer so need to reload
  Idle_Cmd();
  Write_FIFO(My_Key1);
  Write_FIFO(My_Key2);
  Write_FIFO(My_Key3);
  Write_FIFO(My_Key4);
  Write_FIFO(My_Key5);
  Write_FIFO(My_Key6); // end key A
  Write_FIFO(0xFF); // write access flags
  Write_FIFO(0x07);
  Write_FIFO(0x80);
  Write_FIFO(0x69); // don't care byte
  Write_FIFO(My_Key1);
  Write_FIFO(My_Key2);
  Write_FIFO(My_Key3);
  Write_FIFO(My_Key4);
  Write_FIFO(My_Key5);
  Write_FIFO(My_Key6); // end key B
  Write_FIFO(CRC_L);
  Write_FIFO(CRC_H);
  Transceive();
  Flush_RX();
  delay(10);
  Read_FIFO();
  temp = Serial.read();
  if (temp != 10) // acknowledge
    Bad_Write();
}

// Called to initialize a new tag or one with corrupted data
void Init_Tag()
{
  Write_Tag();
  Millions = 0;
  Ten_Thou = 0;
  Hundreds = 0;
  Ones = 0;
  Send_Data();
  Tag_Blk_Num = Trail_Blk;
  Write_Tag();
  Send_Keys();
  Tag_Done();
}

// Gets the credit value from the tag and verifies it against the
// complemented copy.  Returns '1' if valid or '0' if invalid.
void Get_Credit()
{
  Read_Tag();
  Flush_RX();
  Read_FIFO();
  Read_FIFO();
  Read_FIFO();
  Read_FIFO();
  Read_FIFO();
  Read_FIFO();
  Read_FIFO();
  Read_FIFO();
  delay(10);
  Millions = Serial.read();
  Ten_Thou = Serial.read();
  Hundreds = Serial.read();
  Ones = Serial.read();
  // Read 1's complement version
  Comp_4 = ~Serial.read();
  Comp_3 = ~Serial.read();
  Comp_2 = ~Serial.read();
  Comp_1 = ~Serial.read();
  // verify values
  temp = Millions - Comp_4 + Ten_Thou - Comp_3 + Hundreds - Comp_2 + Ones - Comp_1;
  if (temp != 0)
    Bad_Credit = true;
  else
    Bad_Credit = false;
}

void Get_Backup()
{
  Tag_Blk_Num = Bkup_Blk;
  Read_Tag();
  Get_Credit();
  if (Bad_Credit == true)
    Bad_Card();
}

void Match_Fail() {
  Bad_Data();
  Get_Backup();
  Tag_Blk_Num = Data_Blk;
  Write_Tag();
  Send_Data();
  Read_Tag();
  Get_Credit();

  if (Bad_Credit == true)
    Bad_Card();

  Data_Fixed();
  Disp_Credit();
}

void Inc_Credit() {
  Ones = Ones + Inc1;
  Hundreds = Hundreds + Inc100;
  Ten_Thou = Ten_Thou + Inc10k;
  Millions = Millions + Inc1M;

  if (((Ones & 0x0F) > 9) || ((Ones & 0x0F) < (Inc1 & 0x0F)))
    Ones = Ones + 6;
  if (((Ones & 0xF0) > 0x90) || ((Ones & 0xF0) < (Inc1 & 0xF0))) {
    Ones = Ones + 0x60;
    Hundreds++;
  }

  if (((Hundreds & 0x0F) > 9) || ((Hundreds & 0x0F) < (Inc100 & 0x0F)))
    Hundreds = Hundreds + 6;
  if (((Hundreds & 0xF0) > 0x90) || ((Hundreds & 0xF0) < (Inc100 & 0xF0))) {
    Hundreds = Hundreds + 0x60;
    Ten_Thou++;
  }

  if (((Ten_Thou & 0x0F) > 9) || ((Ten_Thou & 0x0F) < (Inc10k & 0x0F)))
    Ten_Thou = Ten_Thou + 6;
  if (((Ten_Thou & 0xF0) > 0x90) || ((Ten_Thou & 0xF0) < (Inc10k & 0xF0))) {
    Ten_Thou = Ten_Thou + 0x60;
    Millions++;
  }

  if (((Millions & 0x0F) > 9) || ((Millions & 0x0F) < (Inc1M & 0x0F)))
    Millions = Millions + 6;
  if (((Millions & 0xF0) > 0x90) || ((Millions & 0xF0) < (Inc1M & 0xF0)))
    Full_Credit();
  else {
    Tag_Blk_Num = Bkup_Blk;
    Write_Tag();
    Send_Data();
    Tag_Blk_Num = Data_Blk;
    Write_Tag();
    Send_Data();
    Read_Tag();
    Get_Credit();

    if (Bad_Credit == true)
      Match_Fail();
    else
      Disp_Credit();
  }
}

void Dec_Credit() {
  if ((Ones & 0x0F) < (Dec1 & 0x0F))
    Ones = Ones - 6;
  if ((Ones & 0xF0) < (Dec1 & 0xF0)) {
    Ones = Ones - 0x60;
    Hundreds--;
  }

  if ((Hundreds & 0x0F) < (Dec100 & 0x0F))
    Hundreds = Hundreds - 6;
  if ((Hundreds & 0xF0) < (Dec100 & 0xF0)) {
    Hundreds = Hundreds - 0x60;
    Ten_Thou--;
  }

  if ((Ten_Thou & 0x0F) < (Dec10k & 0x0F))
    Ten_Thou = Ten_Thou - 6;
  if ((Ten_Thou & 0xF0) < (Dec10k & 0xF0)) {
    Ten_Thou = Ten_Thou - 0x60;
    Millions--;
  }

  if ((Millions & 0x0F) < (Dec1M & 0x0F))
    Millions = Millions - 6;
  if ((Millions & 0xF0) < (Dec1M & 0xF0))
    Low_Credit();
  else {
    Ones = Ones - Dec1;
    Hundreds = Hundreds - Dec100;
    Ten_Thou = Ten_Thou - Dec10k;
    Millions = Millions - Dec1M;
    if ((Millions & 0xF0) > 0x90)
      Low_Credit();
    else {
      Tag_Blk_Num = Bkup_Blk;
      Write_Tag();
      Send_Data();
      Tag_Blk_Num = Data_Blk;
      Write_Tag();
      Send_Data();
      Read_Tag();
      Get_Credit();

      if (Bad_Credit == true)
        Match_Fail();
      else
        Disp_Credit();
    }
  }
}

// ---------------------------------------------------
// Display Messages
// ---------------------------------------------------

void No_Resp()
{
  Clear_LCD();
  Send_Byte('N');
  Send_Byte('O');
  Send_Byte(' ');
  Send_Byte('R');
  Send_Byte('E');
  Send_Byte('S');
  Send_Byte('P');
  Send_Byte('O');
  Send_Byte('N');
  Send_Byte('S');
  Send_Byte('E');
  Error_Flg = true;
  delay(2000);
}

void Bad_Resp()
{
  Clear_LCD();
  Send_Byte('B');
  Send_Byte('A');
  Send_Byte('D');
  Send_Byte(' ');
  Send_Byte('R');
  Send_Byte('E');
  Send_Byte('S');
  Send_Byte('P');
  Send_Byte('O');
  Send_Byte('N');
  Send_Byte('S');
  Send_Byte('E');
  LCD_Line2();
  Send_Byte('E');
  Send_Byte('X');
  Send_Byte('P');
  Send_Byte('E');
  Send_Byte('C');
  Send_Byte('T');
  Send_Byte(':');
  Make_ASCII(Exp_Resp);
  Send_Byte(' ');
  Send_Byte('G');
  Send_Byte('O');
  Send_Byte('T');
  Send_Byte(':');
  Make_ASCII(Cmd_Resp);
  Error_Flg = true;
  delay(2000);
}

void Scan_Msg() {
  Clear_LCD();
  Send_Byte('R');
  Send_Byte('E');
  Send_Byte('A');
  Send_Byte('D');
  Send_Byte('Y');
  Send_Byte(' ');
  Send_Byte('T');
  Send_Byte('O');
  Send_Byte(' ');
  Send_Byte('S');
  Send_Byte('C');
  Send_Byte('A');
  Send_Byte('N');
  delay(1000);
}

void Bad_Scan() {
  Clear_LCD();
  Send_Byte('B');
  Send_Byte('A');
  Send_Byte('D');
  Send_Byte(' ');
  Send_Byte('S');
  Send_Byte('C');
  Send_Byte('A');
  Send_Byte('N');
  Error_Flg = true;
  delay(2000);
}

void Select_Msg() {
  Clear_LCD();
  Send_Byte('T');
  Send_Byte('A');
  Send_Byte('G');
  Send_Byte(' ');
  Send_Byte('S');
  Send_Byte('E');
  Send_Byte('L');
  Send_Byte('E');
  Send_Byte('C');
  Send_Byte('T');
  Send_Byte('E');
  Send_Byte('D');
  delay(1000);
}

void Bad_Read() {
  Clear_LCD();
  Send_Byte('B');
  Send_Byte('A');
  Send_Byte('D');
  Send_Byte(' ');
  Send_Byte('R');
  Send_Byte('E');
  Send_Byte('A');
  Send_Byte('D');
  Error_Flg = true;
  delay(2000);
}

void Bad_Write() {
  Clear_LCD();
  Send_Byte('B');
  Send_Byte('A');
  Send_Byte('D');
  Send_Byte(' ');
  Send_Byte('W');
  Send_Byte('R');
  Send_Byte('I');
  Send_Byte('T');
  Send_Byte('E');
  Error_Flg = true;
  delay(2000);
}

void Disp_UID() {
  Clear_LCD();
  Send_Byte('U');
  Send_Byte('I');
  Send_Byte('D');
  Send_Byte(':');
  Send_Byte(' ');
  Make_ASCII(UID1);
  Make_ASCII(UID2);
  Make_ASCII(UID3);
  Make_ASCII(UID4);
  delay(2000);
}

void Auth_Msg() {
  Clear_LCD();
  Send_Byte('A');
  Send_Byte('U');
  Send_Byte('T');
  Send_Byte('H');
  Send_Byte('E');
  Send_Byte('N');
  Send_Byte('T');
  Send_Byte('I');
  Send_Byte('C');
  Send_Byte('A');
  Send_Byte('T');
  Send_Byte('E');
  Send_Byte('D');
  delay(2000);
}

void Bad_Card() {
  Clear_LCD();
  Send_Byte('B');
  Send_Byte('A');
  Send_Byte('D');
  Send_Byte(' ');
  Send_Byte('C');
  Send_Byte('A');
  Send_Byte('R');
  Send_Byte('D');
  Error_Flg = true;
  Reinit_Tag = true;
  delay(2000);
}

void Bad_Data() {
  Clear_LCD();
  Send_Byte('D');
  Send_Byte('A');
  Send_Byte('T');
  Send_Byte('A');
  Send_Byte(' ');
  Send_Byte('C');
  Send_Byte('O');
  Send_Byte('R');
  Send_Byte('R');
  Send_Byte('U');
  Send_Byte('P');
  Send_Byte('T');
  Send_Byte(':');
  LCD_Line2();
  Send_Byte('C');
  Send_Byte('H');
  Send_Byte('E');
  Send_Byte('C');
  Send_Byte('K');
  Send_Byte('I');
  Send_Byte('N');
  Send_Byte('G');
  Send_Byte(' ');
  Send_Byte('B');
  Send_Byte('A');
  Send_Byte('C');
  Send_Byte('K');
  Send_Byte('U');
  Send_Byte('P');
  delay(3000);
}

void Low_Credit() {
  Clear_LCD();
  Send_Byte('C');
  Send_Byte('R');
  Send_Byte('E');
  Send_Byte('D');
  Send_Byte('I');
  Send_Byte('T');
  Send_Byte('S');
  Send_Byte(' ');
  Send_Byte('T');
  Send_Byte('O');
  Send_Byte('O');
  Send_Byte(' ');
  Send_Byte('L');
  Send_Byte('O');
  Send_Byte('W');
  Error_Flg = true;
  delay(2000);
}

void Full_Credit() {
  Clear_LCD();
  Send_Byte('W');
  Send_Byte('A');
  Send_Byte('L');
  Send_Byte('L');
  Send_Byte('E');
  Send_Byte('T');
  Send_Byte(' ');
  Send_Byte('F');
  Send_Byte('U');
  Send_Byte('L');
  Send_Byte('L');
  Error_Flg = true;
  delay(2000);
}

void Disp_Credit() {
  Clear_LCD();
  Send_Byte('C');
  Send_Byte('R');
  Send_Byte('E');
  Send_Byte('D');
  Send_Byte('I');
  Send_Byte('T');
  Send_Byte('S');
  Send_Byte(':');
  LCD_Line2();
  Find_First();
  delay(2000);
}

void Tag_Done() {
  Clear_LCD();
  Send_Byte('T');
  Send_Byte('A');
  Send_Byte('G');
  Send_Byte(' ');
  Send_Byte('I');
  Send_Byte('N');
  Send_Byte('I');
  Send_Byte('T');
  Send_Byte('I');
  Send_Byte('A');
  Send_Byte('L');
  Send_Byte('I');
  Send_Byte('Z');
  Send_Byte('E');
  Send_Byte('D');
  delay(2000);
}

void Data_Fixed() {
  Clear_LCD();
  Send_Byte('D');
  Send_Byte('A');
  Send_Byte('T');
  Send_Byte('A');
  Send_Byte(' ');
  Send_Byte('R');
  Send_Byte('E');
  Send_Byte('P');
  Send_Byte('A');
  Send_Byte('I');
  Send_Byte('R');
  Send_Byte('E');
  Send_Byte('D');
  delay(2000);
}

void Done_Msg() {
  Clear_LCD();
  Send_Byte('R');
  Send_Byte('E');
  Send_Byte('M');
  Send_Byte('O');
  Send_Byte('V');
  Send_Byte('E');
  Send_Byte(' ');
  Send_Byte('T');
  Send_Byte('A');
  Send_Byte('G');
  delay(2000);
}
